package gomatrixbot

import (
	"context"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"time"

	"maunium.net/go/mautrix/id"
)

type geoNetResponse struct {
	Type     string `json:"type"`
	Features []struct {
		Type     string `json:"type"`
		Geometry struct {
			Type        string    `json:"type"`
			Coordinates []float64 `json:"coordinates"`
		} `json:"geometry"`
		Properties struct {
			PublicID  string    `json:"publicID"`
			Time      time.Time `json:"time"`
			Depth     float64   `json:"depth"`
			Magnitude float64   `json:"magnitude"`
			Mmi       int       `json:"mmi"`
			Locality  string    `json:"locality"`
			Quality   string    `json:"quality"`
		} `json:"properties"`
	} `json:"features"`
}

func (mtrx *MtrxClient) checkForQuakes(ctx context.Context, roomID id.RoomID, intensity string) {
	// Get JSON Weather
	url := "https://api.geonet.org.nz/quake?"
	httpClient := http.Client{Timeout: time.Second * 2}
	req, err := http.NewRequest(http.MethodGet, url, nil)
	if err != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching geonet feed: "+err.Error())
		return
	}

	q := req.URL.Query()
	q.Add("MMI", intensity)
	req.URL.RawQuery = q.Encode()

	req.Header.Set("Accept", "application/vnd.geo+json;version=2")
	res, getErr := httpClient.Do(req)
	if getErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching geonet feed: "+getErr.Error())
		return
	}

	if res.Body != nil {
		defer res.Body.Close()
	}

	body, readErr := io.ReadAll(res.Body)
	if readErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching geonet feed: "+readErr.Error())
		return
	}

	// Decode into weatherResponse struct
	geonet := geoNetResponse{}
	jsonErr := json.Unmarshal(body, &geonet)
	if jsonErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching geonet: "+jsonErr.Error())
		return
	}

	msg := "Last 5 quakes with intensity " + fmt.Sprint(intensity) + "+:\n\n"
	i := 0
	loc, err := time.LoadLocation("Pacific/Auckland")
	if err != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching geonet: "+err.Error())
		return
	}

	for _, geo := range geonet.Features {
		msg += fmt.Sprintf(
			"%s: %.1fM at %s\n",
			geo.Properties.Time.In(loc).Format(time.DateTime),
			geo.Properties.Magnitude,
			geo.Properties.Locality,
		)
		i++
		if i >= 5 {
			break
		}
	}

	mtrx.sendMessage(ctx, roomID, msg)
}
