package gomatrixbot

import (
	"context"
	"errors"
	"fmt"
	"gomatrixbot/internal/database"
	"image"
	"os"
	"os/exec"
	"sort"
	"strings"
	"sync"
	"time"

	"github.com/chzyer/readline"
	_ "github.com/mattn/go-sqlite3"
	"github.com/peterhellberg/flip"
	"github.com/rs/zerolog"
	"github.com/rwtodd/Go.Sed/sed"
	"go.mau.fi/util/exzerolog"
	"maunium.net/go/mautrix"
	"maunium.net/go/mautrix/crypto/cryptohelper"
	"maunium.net/go/mautrix/event"
	"maunium.net/go/mautrix/id"
)

var (
	// Matrix creds
	MatrixHost     string
	MatrixUsername string
	MatrixPassword string
	MautrixDB      = BASE_PATH + "/mautrix.db"

	mtrx MtrxClient
)

type MtrxClient struct {
	c         *mautrix.Client
	startTime int64
	adminRoom id.RoomID
	db        *database.Database

	savedFiles     map[id.EventID]string
	events         map[id.EventID]string
	recentMessages map[id.RoomID]map[int64]string
	pics           map[int64]string

	// This is gonna be a mem leak... To fix I guess
	quoteCache map[id.EventID]*Quote

	quitMeDaddy chan struct{}
}

// Run - starts bot!
func Run(db *database.Database) {
	mtrx := MtrxClient{}
	mtrx.db = db
	mtrx.startTime = time.Now().UnixMilli()

	mtrx.recentMessages = make(map[id.RoomID]map[int64]string)
	mtrx.adminRoom = id.RoomID(os.Getenv("MATRIX_ADMIN_ROOM"))

	pics, err := mtrx.db.GetAllPics()
	if err != nil {
		fmt.Println(err)
		mtrx.pics = make(map[int64]string, 0)
	} else {
		mtrx.pics = pics
	}

	// boot matrix
	client, err := mautrix.NewClient(MatrixHost, "", "")
	if err != nil {
		panic(err)
	}
	mtrx.c = client

	rl, err := readline.New("[no room]> ")
	if err != nil {
		panic(err)
	}
	defer rl.Close()

	log := zerolog.New(zerolog.NewConsoleWriter(func(w *zerolog.ConsoleWriter) {
		w.Out = rl.Stdout()
		w.TimeFormat = time.Stamp
	})).With().Timestamp().Logger()
	log = log.Level(zerolog.InfoLevel)

	exzerolog.SetupDefaults(&log)
	mtrx.c.Log = log

	// Purge old files and init maps
	purgeLog := mtrx.purgeOutput()
	log.Info().Msg(purgeLog)

	var lastRoomID id.RoomID
	syncer := mtrx.c.Syncer.(*mautrix.DefaultSyncer)
	syncer.OnEventType(event.EventMessage, func(ctx context.Context, evt *event.Event) {
		lastRoomID = evt.RoomID
		rl.SetPrompt(fmt.Sprintf("%s> ", lastRoomID))
		log.Info().
			Str("sender", evt.Sender.String()).
			Str("type", evt.Type.String()).
			Str("body", evt.Content.AsMessage().Body)

		// Thread for s p e e d
		go mtrx.handleMessageEvent(ctx, evt)
	})

	syncer.OnEventType(event.EventReaction, func(ctx context.Context, evt *event.Event) {
		lastRoomID = evt.RoomID
		rl.SetPrompt(fmt.Sprintf("%s> ", lastRoomID))
		log.Info().
			Str("sender", evt.Sender.String()).
			Str("type", evt.Type.String()).
			Str("body", evt.Content.AsMessage().Body)

		// Thread for s p e e d
		go mtrx.handleReaction(ctx, evt)
	})
	syncer.OnEventType(event.StateMember, func(ctx context.Context, evt *event.Event) {
		if evt.GetStateKey() == mtrx.c.UserID.String() && evt.Content.AsMember().Membership == event.MembershipInvite {
			_, err := mtrx.c.JoinRoomByID(ctx, evt.RoomID)
			if err == nil {
				lastRoomID = evt.RoomID
				rl.SetPrompt(fmt.Sprintf("%s> ", lastRoomID))
				log.Info().
					Str("room_id", evt.RoomID.String()).
					Str("inviter", evt.Sender.String()).
					Msg("Joined room after invite")
			} else {
				log.Error().Err(err).
					Str("room_id", evt.RoomID.String()).
					Str("inviter", evt.Sender.String()).
					Msg("Failed to join room after invite")
			}
		}
	})

	cryptoHelper, err := cryptohelper.NewCryptoHelper(mtrx.c, []byte("meow"), MautrixDB)
	if err != nil {
		panic(err)
	}

	cryptoHelper.LoginAs = &mautrix.ReqLogin{
		Type:       mautrix.AuthTypePassword,
		Identifier: mautrix.UserIdentifier{Type: mautrix.IdentifierTypeUser, User: MatrixUsername},
		Password:   MatrixPassword,
	}
	err = cryptoHelper.Init(context.TODO())
	if err != nil {
		panic(err)
	}
	// Set the client crypto helper in order to automatically encrypt outgoing messages
	mtrx.c.Crypto = cryptoHelper

	log.Info().Msg("Now running")
	syncCtx, cancelSync := context.WithCancel(context.Background())
	var syncStopWait sync.WaitGroup
	syncStopWait.Add(1)

	go func() {
		err = mtrx.c.SyncWithContext(syncCtx)
		defer syncStopWait.Done()
		if err != nil && !errors.Is(err, context.Canceled) {
			panic(err)
		}
	}()

	// end pic checker

	for {
		line, err := rl.Readline()
		if err != nil { // io.EOF
			break
		}
		if lastRoomID == "" {
			log.Error().Msg("Wait for an incoming message before sending messages")
			continue
		}
		resp, err := mtrx.c.SendText(context.TODO(), lastRoomID, line)
		if err != nil {
			log.Error().Err(err).Msg("Failed to send event")
		} else {
			log.Info().Str("event_id", resp.EventID.String()).Msg("Event sent")
		}
	}

	if mtrx.adminRoom != "" {
		// mtrx.c.SendText(context.Background(), mtrx.adminRoom, "Hello world.")
	}

	// Keep it running!!
	mtrx.quitMeDaddy = make(chan struct{})
	ticker := time.NewTicker(1 * time.Minute)
	go func() {
		for {
			select {
			case <-ticker.C:
				// Pic checker
				mtrx.UpdatePics()
			case <-mtrx.quitMeDaddy:
				log.Info().Msg("Received quit command!!!")
				cancelSync()
				syncStopWait.Wait()
				err = cryptoHelper.Close()
				if err != nil {
					log.Error().Err(err).Msg("Error closing database")
				}
				os.Exit(0)
			}
		}
	}()
}

func (mtrx *MtrxClient) setTopic(ctx context.Context, evt *event.Event) {
	if val, ok := mtrx.quoteCache[evt.Content.AsReaction().GetRelatesTo().EventID]; ok {
		quote := val.Quote

		if strings.HasPrefix(val.Quote, ">") && strings.Contains(val.Quote, "\n\n") {
			parts := strings.Split(val.Quote, "\n\n")
			quote = strings.Join(parts[1:], "\n")
		}

		if quote != "" {
			mtrx.c.Log.Info().Msg("Setting qutoe")
			_, err := mtrx.c.SendStateEvent(ctx, evt.RoomID, event.StateTopic, "", map[string]interface{}{"topic": quote})
			if err != nil {
				mtrx.c.Log.Error().Err(err).Msg("Failed to set topic")
			}
		}
	}
}

func (mtrx *MtrxClient) handleMessageEvent(ctx context.Context, evt *event.Event) {
	if _, ok := mtrx.events[evt.ID]; !ok {
		mtrx.events[evt.ID] = evt.Content.AsMessage().Body
	} else {
		return
	}

	// If syncing older messages.. stop that now
	if evt.Timestamp < mtrx.startTime {
		return
	}

	// If it's an edit, use the edited text.
	msg := evt.Content.AsMessage().Body
	if evt.Content.AsMessage().NewContent != nil {
		msg = evt.Content.AsMessage().NewContent.Body
	}

	// If parsing own messages.. stop that too
	if evt.Sender.String() == mtrx.c.UserID.String() &&
		!strings.HasPrefix(msg, ".gif") &&
		!strings.HasPrefix(msg, ".giphy") &&
		!strings.HasPrefix(msg, ".xkcd") &&
		!strings.HasPrefix(msg, ".b") &&
		!strings.HasPrefix(msg, ".r") &&
		!strings.HasPrefix(msg, ".rb") {
		return
	}

	// Store in cache if it's not a sed cmd
	if msg != "" && !strings.HasPrefix(msg, "s/") {
		mtrx.quoteCache[evt.ID] = &Quote{
			Quote: msg,
			User:  evt.Sender,
		}

		if _, ok := mtrx.recentMessages[evt.RoomID]; !ok {
			mtrx.recentMessages[evt.RoomID] = make(map[int64]string, 0)
		} else if len(mtrx.recentMessages[evt.RoomID]) > 50 {
			// only keep 50 messages, delete older than 2 hours
			for k := range mtrx.recentMessages[evt.RoomID] {
				if k < time.Now().Add(-time.Hour*2).UnixMicro() {
					delete(mtrx.recentMessages[evt.RoomID], k)
				}
			}
		}

		mtrx.recentMessages[evt.RoomID][time.Now().UnixMicro()] = msg
	}

	// Parse incoming command
	if strings.HasPrefix(msg, ".") {
		mtrx.handleCommand(ctx, evt)
	} else if strings.HasPrefix(msg, "s/") {
		// LETS REGEX THE FUXK OUT OF THIS YOOOO
		engine, err := sed.New(strings.NewReader(msg))
		if err != nil {
			mtrx.sendMessage(ctx, evt.RoomID, err.Error())
			return
		}

		// Order map for replacement
		keys := make([]int64, 0, len(mtrx.recentMessages[evt.RoomID]))
		for k := range mtrx.recentMessages[evt.RoomID] {
			keys = append(keys, k)
		}
		sort.Slice(keys, func(i, j int) bool { return keys[i] > keys[j] })

		for _, k := range keys {
			newOutput, err := engine.RunString(mtrx.recentMessages[evt.RoomID][k])
			if err != nil {
				mtrx.c.Log.Error().Err(err).Msg("Failed to run regex")
				continue
			}

			newOutput = strings.TrimRight(newOutput, "\n")

			if newOutput != "" && newOutput != mtrx.recentMessages[evt.RoomID][k] {
				mtrx.sendMessage(
					ctx,
					evt.RoomID,
					newOutput,
				)
				mtrx.recentMessages[evt.RoomID][time.Now().UnixMicro()] = newOutput
				return
			}
		}

		mtrx.sendMessage(ctx, evt.RoomID, "Could not match `"+msg+"` with recent messages")
	} else if evt.Content.AsMessage().MsgType == event.MsgImage || evt.Content.AsMessage().File != nil {
		mtrx.handleImage(ctx, evt)
	} else {
		// 1 in 10 chance of auto reply
		// n := 10
		// if (rand.IntN(n-1) + 1) == 1 {
		// 	keys := make([]int64, 0, len(mtrx.recentMessages[evt.RoomID]))
		// 	for k := range mtrx.recentMessages[evt.RoomID] {
		// 		keys = append(keys, k)
		// 	}
		// 	sort.Slice(keys, func(i, j int) bool { return keys[i] > keys[j] })

		// 	i := 0
		// 	context := []string{}
		// 	for _, k := range keys {
		// 		if i >= 4 {
		// 			mtrx.aiQuery(ctx, evt.RoomID, context)
		// 			return
		// 		}

		// 		context = append(context, mtrx.recentMessages[evt.RoomID][k])
		// 		i = i + 1
		// 	}
		// }

		// if using tbqh..
		if strings.Contains(msg, "tbqh") {
			mtrx.PostTenorGif(ctx, evt.RoomID, []string{"banned"}, 20)
			return
		}

		// if using tbqh..
		// if strings.Contains(msg, "gary") {
		// 	mtrx.postBirbWithText(ctx, evt, mtrx.getAnyRandomQuote(true), 0)
		// 	return
		// }

		// If tagged gary, post guilty
		// if evt.Content.AsMessage().Mentions != nil {
		// 	if len(evt.Content.AsMessage().Mentions.UserIDs) > 0 {
		// 		for _, v := range evt.Content.AsMessage().Mentions.UserIDs {
		// 			if v == "@gary:possum.lol" {
		// 				mtrx.PostTenorGif(ctx, evt.RoomID, []string{"wasn't me"}, 20)
		// 				return
		// 			}
		// 		}
		// 	}
		// }
	}
}

func (mtrx *MtrxClient) FlipText(ctx context.Context, roomID id.RoomID) {
	keys := make([]int64, 0, len(mtrx.recentMessages[roomID]))
	for k := range mtrx.recentMessages[roomID] {
		keys = append(keys, k)
	}
	sort.Slice(keys, func(i, j int) bool { return keys[i] > keys[j] })

	i := 0
	for _, k := range keys {
		if i == 0 {
			i++
			continue
		}

		mtrx.sendMessage(ctx, roomID, flip.Table(mtrx.recentMessages[roomID][k]))
		return
	}
}

func (mtrx *MtrxClient) sendMessage(ctx context.Context, roomID id.RoomID, text string) {
	resp, err := mtrx.c.SendText(ctx, roomID, text)
	if err != nil {
		mtrx.c.Log.Error().Err(err).Msg("Failed to send event")
	} else {
		mtrx.c.Log.Info().Str("event_id", resp.EventID.String()).Msg("Event sent")
	}
}

func (mtrx *MtrxClient) sendNotice(ctx context.Context, roomID id.RoomID, text string) {
	resp, err := mtrx.c.SendNotice(ctx, roomID, text)
	if err != nil {
		mtrx.c.Log.Error().Err(err).Msg("Failed to send event")
	} else {
		mtrx.c.Log.Info().Str("event_id", resp.EventID.String()).Msg("Event sent")
	}
}

func (mtrx *MtrxClient) uploadAndPostFile(ctx context.Context, roomID id.RoomID, newLocalFile string) {
	// Read replaced file
	b, err := os.ReadFile(newLocalFile) // just pass the file name
	if err != nil {
		mtrx.c.Log.Err(err).Msg("Failed to read new file")
		return
	}

	fileParts := strings.Split(newLocalFile, ".")
	fileType := fileParts[len(fileParts)-1]
	mime := "image/png"
	if fileType == "jpg" {
		mime = "image/jpeg"
	} else if fileType == "mp4" {
		mime = "video/mp4"
	}

	media, err := mtrx.c.UploadBytesWithName(ctx, b, mime, "face."+fileType)
	if err != nil {
		mtrx.c.Log.Err(err).Msg("Failed to upload file")
		return
	}
	resp, err := mtrx.c.SendMessageEvent(ctx, roomID, event.EventMessage, &event.MessageEventContent{
		MsgType:  event.MsgImage,
		FileName: "face." + fileType,
		Body:     "",
		Info: &event.FileInfo{
			MimeType: mime,
		},
		URL: id.ContentURIString(media.ContentURI.String()),
	})

	if err != nil {
		mtrx.c.Log.Err(err).Msg("Failed to send event")
		return
	}

	mtrx.savedFiles[resp.EventID] = newLocalFile
}

func (mtrx *MtrxClient) uploadAndPostVideo(ctx context.Context, roomID id.RoomID, filename string) {
	// Read replaced file
	b, err := os.ReadFile(filename) // just pass the file name
	if err != nil {
		mtrx.c.Log.Err(err).Msg("Failed to read new file")
		return
	}

	width := 640
	height := 480

	// Gen thumbnail
	var thumb *mautrix.RespMediaUpload
	cmd := exec.Command("/usr/bin/ffmpeg", "-i", filename, "-ss", "00:00:01.000", "-vframes", "1", filename+".png")
	cmd.Dir = "/tmp"
	err = cmd.Run()
	if err != nil {
		mtrx.c.Log.Err(err).Msg("Failed to get thumbnail")
	} else {
		b2, err := os.ReadFile(filename + ".png") // just pass the file name
		if err != nil {
			mtrx.c.Log.Err(err).Msg("Failed to read new file")
			return
		}

		thumb, err = mtrx.c.UploadBytesWithName(ctx, b2, "image/png", "thumbnail.png")
		if err != nil {
			mtrx.c.Log.Err(err).Msg("Failed to upload file")
		}

		// Get dimensions if set
		if reader, err := os.Open(filename + ".png"); err == nil {
			defer reader.Close()
			im, _, err := image.DecodeConfig(reader)
			if err != nil {
				mtrx.c.Log.Err(err).Msg("Failed to get image dimensions")
			} else {
				width = im.Width
				height = im.Height
			}
		}
	}

	// Upload
	media, err := mtrx.c.UploadBytesWithName(ctx, b, "video/mp4", "video.mp4")
	if err != nil {
		mtrx.sendMessage(ctx, roomID, "Failed to upload video: "+err.Error())
		mtrx.c.Log.Err(err).Msg("Failed to upload file")
		return
	}

	// Get filesize
	size := int64(0)
	fi, err := os.Stat(filename)
	if err == nil {
		// get the size
		size = fi.Size()
	}

	evtContent := &event.MessageEventContent{
		MsgType: event.MsgVideo,
		Info: &event.FileInfo{
			MimeType: "video/mp4",
			Width:    width,
			Height:   height,
		},
		FileName: "video.mp4",
		Body:     "",
		URL:      id.ContentURIString(media.ContentURI.String()),
	}

	// Add thumb if set
	if thumb != nil {
		evtContent.Info.ThumbnailURL = id.ContentURIString(thumb.ContentURI.String())
	}

	// Add size if not zero
	if size != 0 {
		evtContent.Info.Size = int(size)
	}

	_, err = mtrx.c.SendMessageEvent(ctx, roomID, event.EventMessage, evtContent)

	if err != nil {
		mtrx.c.Log.Err(err).Msg("Failed to send event")
		return
	}
}

func (mtrx *MtrxClient) purge(ctx context.Context, room id.RoomID) {
	// get events
	msgs, err := mtrx.c.Messages(ctx, room, "", "", mautrix.DirectionForward, nil, 50)
	if err != nil {
		mtrx.sendMessage(ctx, room, "Failed to get messages: "+err.Error())
		return
	}

	// mtrx.sendMessage(ctx, room, "Purging messages..."+msgs.Start+": "+msgs.End)
	for _, v := range msgs.Chunk {
		if v.Type != event.EventMessage {
			continue
		}

		// mtrx.sendMessage(ctx, room, "Redacting event: "+v.ID.String())
		_, err = mtrx.c.RedactEvent(ctx, room, v.ID)
		if err != nil {
			mtrx.sendMessage(ctx, room, "Failed to redact event: "+err.Error())
			return
		}
	}

	for {
		// Break if finished!
		if msgs.End == "" {
			break
		}

		msgs, err = mtrx.c.Messages(ctx, room, msgs.End, "", mautrix.DirectionForward, nil, 50)
		if err != nil {
			mtrx.sendMessage(ctx, room, "Failed to get messages: "+err.Error())
			return
		}

		// mtrx.sendMessage(ctx, room, "Purging messages..."+msgs.Start+": "+msgs.End)

		for _, v := range msgs.Chunk {
			if v.Type != event.EventMessage {
				continue
			}

			_, err = mtrx.c.RedactEvent(ctx, room, v.ID)
			if err != nil {
				mtrx.sendMessage(ctx, room, "Failed to redact event: "+err.Error())
				return
			}
		}

		// time.Sleep(1 * time.Second)
	}

	mtrx.sendMessage(ctx, room, "Hello frens")
}

// Start string         `json:"start"`
// Chunk []*event.Event `json:"chunk"`
// State []*event.Event `json:"state"`
// End   string         `json:"end,omitempty"`
