package gomatrixbot

import (
	"context"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"os"
	"time"

	"maunium.net/go/mautrix/id"
)

type weatherGeoResponse []struct {
	Name    string  `json:"name"`
	Lat     float64 `json:"lat"`
	Lon     float64 `json:"lon"`
	Country string  `json:"country"`
	State   string  `json:"state,omitempty"`
}

type weatherResponse struct {
	Weather []struct {
		ID          int    `json:"id"`
		Main        string `json:"main"`
		Description string `json:"description"`
		Icon        string `json:"icon"`
	} `json:"weather"`
	Base string `json:"base"`
	Main struct {
		Temp      float64 `json:"temp"`
		FeelsLike float64 `json:"feels_like"`
		TempMin   float64 `json:"temp_min"`
		TempMax   float64 `json:"temp_max"`
		Pressure  int     `json:"pressure"`
		Humidity  int     `json:"humidity"`
	} `json:"main"`
	Visibility int `json:"visibility"`
	Wind       struct {
		Speed float64 `json:"speed"`
		Deg   int     `json:"deg"`
	} `json:"wind"`
	Rain struct {
		OneHour float64 `json:"1h"`
	} `json:"rain"`
	Clouds struct {
		All int `json:"all"`
	} `json:"clouds"`
	Dt   int    `json:"dt"`
	Name string `json:"name"`
}

type weatherForecastResponse struct {
	Cnt  int `json:"cnt"`
	List []struct {
		Dt   int `json:"dt"`
		Main struct {
			Temp      float64 `json:"temp"`
			FeelsLike float64 `json:"feels_like"`
			TempMin   float64 `json:"temp_min"`
			TempMax   float64 `json:"temp_max"`
			Pressure  int     `json:"pressure"`
			Humidity  int     `json:"humidity"`
		} `json:"main"`
		Weather []struct {
			ID          int    `json:"id"`
			Main        string `json:"main"`
			Description string `json:"description"`
			Icon        string `json:"icon"`
		} `json:"weather"`
		Clouds struct {
			All int `json:"all"`
		} `json:"clouds"`
		Wind struct {
			Speed float64 `json:"speed"`
			Deg   int     `json:"deg"`
			Gust  float64 `json:"gust"`
		} `json:"wind"`
		Visibility int     `json:"visibility"`
		Pop        float64 `json:"pop"`
		Rain       struct {
			ThreeH float64 `json:"3h"`
		} `json:"rain,omitempty"`
		DtTxt string `json:"dt_txt"`
	} `json:"list"`
}

// getLonLat fetches the latitude and longitude for a given location
func (mtrx *MtrxClient) getLonLat(ctx context.Context, apiKey string, location string) (weatherGeoResponse, error) {
	weather := weatherGeoResponse{}

	// Get JSON Weather
	url := fmt.Sprintf("http://api.openweathermap.org/geo/1.0/direct?q=%s&limit=1&appid=%s", url.QueryEscape(location), apiKey)
	httpClient := http.Client{Timeout: time.Second * 2}
	req, err := http.NewRequest(http.MethodGet, url, nil)
	if err != nil {
		return weather, err
	}

	req.Header.Set("User-Agent", "GoMatrixBot v1.0")
	res, getErr := httpClient.Do(req)
	if getErr != nil {
		return weather, getErr
	}

	if res.Body != nil {
		defer res.Body.Close()
	}

	body, readErr := io.ReadAll(res.Body)
	if readErr != nil {
		return weather, readErr
	}

	// Decode into weatherResponse struct
	jsonErr := json.Unmarshal(body, &weather)

	return weather, jsonErr
}

// fetchWeather fetches the current weather for a given location
func (mtrx *MtrxClient) fetchWeather(ctx context.Context, roomID id.RoomID, location string) {
	apiKey := os.Getenv("OPENWEATHERMAP_API_KEY")
	if apiKey == "" {
		mtrx.sendMessage(ctx, roomID, "OPENWEATHERMAP_API_KEY not set")
		return
	}

	// Get Lon/Lat
	geo, err := mtrx.getLonLat(ctx, apiKey, location)
	if err != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching geolocation: "+err.Error())
		return
	} else if len(geo) == 0 {
		mtrx.sendMessage(ctx, roomID, "Error fetching geolocation: "+location+" not found.")
		return
	}

	// Get JSON Weather
	url := fmt.Sprintf(
		"https://api.openweathermap.org/data/2.5/weather?lat=%f&lon=%f&units=metric&appid=%s",
		geo[0].Lat,
		geo[0].Lon,
		apiKey,
	)
	fmt.Print(url)

	httpClient := http.Client{Timeout: time.Second * 2}
	req, err := http.NewRequest(http.MethodGet, url, nil)
	if err != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching weather: "+err.Error())
		return
	}

	req.Header.Set("User-Agent", "ezyWeather")
	res, getErr := httpClient.Do(req)
	if getErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching weather: "+getErr.Error())
		return
	}

	if res.Body != nil {
		defer res.Body.Close()
	}

	body, readErr := io.ReadAll(res.Body)
	if readErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching weather: "+err.Error())
		return
	}

	// Decode into weatherResponse struct
	weather := weatherResponse{}
	jsonErr := json.Unmarshal(body, &weather)
	if jsonErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching weather: "+jsonErr.Error()+"\nIs "+location+" even a real place you muppet?")
		return
	}

	windSpeed := fmt.Sprintf("%.2f", (weather.Wind.Speed * 3.6))
	msg := fmt.Sprintf(
		"%s, %s\n"+
			"Cond: %s - %s | Wind: %vkm/h @ %v\n"+
			"Temp (C): Current: %v° | Feels Like: %v° | Min: %v°, Max: %v°.\n"+
			"Pressure: %vhPa | Humidity: %v%% | Rain: %.1f mm/hr",
		geo[0].Name, geo[0].Country,
		weather.Weather[0].Main, weather.Weather[0].Description, windSpeed, weather.Wind.Deg,
		weather.Main.Temp, weather.Main.FeelsLike, weather.Main.TempMin, weather.Main.TempMax,
		weather.Main.Pressure, weather.Main.Humidity, weather.Rain.OneHour,
	)

	mtrx.sendMessage(ctx, roomID, msg)
}

// featchForecast fetches the weather forecast for a given location
func (mtrx *MtrxClient) fetchForecast(ctx context.Context, roomID id.RoomID, location string) {
	apiKey := os.Getenv("OPENWEATHERMAP_API_KEY")
	if apiKey == "" {
		mtrx.sendMessage(ctx, roomID, "OPENWEATHERMAP_API_KEY not set")
		return
	}

	// Get Lon/Lat
	geo, err := mtrx.getLonLat(ctx, apiKey, location)
	if err != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching geolocation: "+err.Error())
		return
	} else if len(geo) == 0 {
		mtrx.sendMessage(ctx, roomID, "Error fetching geolocation: "+location+" not found.")
		return
	}

	// Get JSON Weather
	url := fmt.Sprintf(
		"https://api.openweathermap.org/data/2.5/forecast?lat=%f&lon=%f&units=metric&appid=%s",
		geo[0].Lat,
		geo[0].Lon,
		apiKey,
	)

	httpClient := http.Client{Timeout: time.Second * 2}
	req, err := http.NewRequest(http.MethodGet, url, nil)
	if err != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching weather: "+err.Error())
		return
	}

	req.Header.Set("User-Agent", "ezyWeather")
	res, getErr := httpClient.Do(req)
	if getErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching weather: "+getErr.Error())
		return
	}

	if res.Body != nil {
		defer res.Body.Close()
	}

	body, readErr := io.ReadAll(res.Body)
	if readErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching weather: "+err.Error())
		return
	}

	// Decode into weatherResponse struct
	weather := weatherForecastResponse{}
	jsonErr := json.Unmarshal(body, &weather)
	if jsonErr != nil {
		mtrx.sendMessage(ctx, roomID, "Error fetching weather: "+jsonErr.Error()+"\nIs "+location+" even a real place you muppet?")
		return
	}

	// Build msg
	loc, err := time.LoadLocation("Pacific/Auckland")
	if err != nil {
		mtrx.sendMessage(ctx, roomID, "Timezone error: "+err.Error())
		return
	}

	i := 0
	max := 16

	msg := fmt.Sprintf("%s, %s Forecast (48 hr)\n", geo[0].Name, geo[0].Country)

	currDay := time.Now().In(loc).Format("02")
	msg = msg + "\n" + time.Now().In(loc).Format("2006/01/02") + ":\n"
	for _, forecast := range weather.List {
		if i > max {
			break
		}

		forcastDate := time.Unix(int64(forecast.Dt), 0).In(loc)
		if forcastDate.Format("02") != currDay {
			msg = msg + "\n" + forcastDate.Format("2006/01/02") + ":\n"
			currDay = forcastDate.Format("02")
		}

		windSpeed := fmt.Sprintf("%.2f", (forecast.Wind.Speed * 3.6))
		rain := ""
		if forecast.Rain.ThreeH != float64(0.0) {
			rain = fmt.Sprintf(" | %.2fmm/3hr", forecast.Rain.ThreeH)
		}
		msg = msg + fmt.Sprintf(
			"%s | %vkm/h@%03d° | %v°C | %dhPa%s\n",
			forcastDate.Format("1504"),
			windSpeed, forecast.Wind.Deg,
			forecast.Main.Temp,
			forecast.Main.Pressure,
			rain,
		)
		i++
	}

	mtrx.sendMessage(ctx, roomID, msg)
}
